/*
 * YOU SHOULD CAREFULLY READ THE FOLLOWING TERMS AND CONDITIONS BEFORE 
 * INSTALLING AND USING THIS PRODUCT, 
 * THE USE OF WHICH IS LICENSED BY 3COM CORPORATION ("3COM") 
 * and others as set forth below for your USE ONLY AS SET FORTH BELOW. 
 * DOWNLOADING, INSTALLING OR OTHERWISE USING ANY PART OF THE SOFTWARE OR 
 * DOCUMENTATION INDICATES THAT YOU ACCEPT THESE TERMS AND CONDITIONS.  
 * IF YOU DO NOT AGREE TO THE TERMS AND CONDITIONS OF THIS AGREEMENT, 
 * DO NOT DOWNLOAD, INSTALL OR OTHERWISE USE THE SOFTWARE OR DOCUMENTATION. 
 * AND IF YOU HAVE RECEIVED THE SOFTWARE AND DOCUMENTATION ON PHYSICAL MEDIA, 
 * RETURN THE ENTIRE PRODUCT WITH THE SOFTWARE AND DOCUMENTATION UNUSED TO THE
 * SUPPLIER WHERE YOU OBTAINED IT.

 * This driver (3c990.c) has been written to work with the 3cr990 product line
 * of network cards, manufactured by 3Com Corp.
 * This driver is not intended for any other product line, including the 3c59x 
 * or 3C90x product lines (although drivers with both of these names,
 * and for both of these product lines, are available ). 
 *
 * It does not work with the 2.3 kernel; only the 2.0 and 2.2 at this point.
 * The driver contains no support now for architectures other than IA-32 bit.
 *
 * To force the media selection, use the command line argument force=X, 
 * where X denotes the selection as follows:
 * 0=10(megabit)Half(Duplex); 1=10Full; 2=100Half; 3=100Full; 4=auto (default)
 * e.g. insmod 3c990.o force=1   
 * ... should give you 10 megabit full duplex forced action.
 * NOTE: Fiber cards are only capable of 100Half and 100Full(default)
 *
 * You may contact 3Com for updates and information (regarding this driver) at:
 * http://support.3com.com/infodeli/tools/nic/3c990.htm
 *
 * You may request or submit driver modifications at:
 * http://support.3com.com/infodeli/tools/nic/linuxrequest.htm
 *
 * If you would like more information about compiling the driver than is 
 * available at the bottom of this file, you may choose to reference:
 * http://cesdis.gsfc.nasa.gov/linux/misc/modules.html
 * and/or you may reference the readme file for the 3c90x driver.
 *
 * License to "tc990image"
 * The binary image "tc990image" is licensed to users by 3Com Corporation, 
 * with intended use for 3Com's Network Interface Card models 3CR990-x.
 * (Where x indicates any extension model number.)
 * LICENSE:  3Com grants you a nonexclusive, nontransferable 
 * (except as specified herein) license to use the tc990image in conjunction 
 * with your use of 3Com's Network Interface Card models 3CR990-x as specified 
 * above. You are not permitted to lease, rent, distribute or sublicense 
 * (except as specified herein) the tc990image or to use the tc990image or in 
 * any other unauthorized manner.  Further, no license is granted to you in the
 * human readable code of the Software (source code).  Except as provided 
 * below, this Agreement does not grant you any rights to patents, copyrights, 
 * trade secrets, trademarks, or any other rights with respect to the Software 
 * or Documentation.
 * Subject to the restrictions set forth herein, the tc990image is licensed to 
 * be used on any workstation or any network server owned by or leased to you, 
 * for your internal use, provided that the tc990image is used only in 
 * connection with this 3Com product.  You may reproduce and provide one (1) 
 * copy of the tc990image for each such workstation or network server on which 
 * the Software is used as permitted hereunder.  Otherwise, the Software and 
 * Documentation may be copied only as essential for backup or archive purposes
 * in support of your use of the Software as permitted hereunder.  Each copy 
 * of the tc990image must contain 3Com's and its licensors' proprietary rights 
 * and copyright notices in the same form as on the original.  You agree not to
 * remove or deface any portion of any legend provided on any licensed program 
 * or documentation delivered to you under this Agreement.
 * ASSIGNMENT;   You may transfer the tc990image and the licenses granted 
 * herein to another party in the same country in which you obtained it if the 
 * other party agrees in writing to accept and be bound by the terms and 
 * conditions of this Agreement. If you transfer the tc990image, you must at 
 * the same time either transfer all copies of the Software and Documentation 
 * to the party or you must destroy any copies not transferred. Except as set 
 * forth above, you may not assign or transfer your rights.
 *
 * NO REVERSE ENGINEERING:  Modification, reverse engineering, reverse 
 * compiling, or disassembly of the tc990image is expressly prohibited. 
 * However, if you are a European Union ("EU") resident, information necessary 
 * to achieve interoperability of the tc990image with other programs within the
 * meaning of the EU Directive on the Legal Protection of Computer Programs is 
 * available to you from 3Com upon written request.
 * TRADE SECRETS; TITLE:  You acknowledge and agree that the structure, 
 * sequence and organization of the tc990image are the valuable trade secrets 
 * of 3Com and its suppliers. You agree to hold such trade secrets in 
 * confidence. You further acknowledge and agree that ownership of, and title 
 * to, the tc990image and all subsequent copies thereof regardless of the form 
 * or media are held by 3Com.
 *
 * License to this driver
 * Some material in this driver has been adopted from pci-skeleton.c, 
 * a Linux PCI network adapter skeleton device driver. 
 * pci-skeleton.c written in 1998-1999 by Donald Becker.
 * The author of pci-skeleton.c may be reached as becker@scyld.com.
 * More information about Becker is available at http://www.scyld.com/
 *
 * Except as otherwise provided, This software and pci-skeleton.c, may be used 
 * and distributed according to the terms of the GNU Public License (GPL), 
 * incorporated herein by reference.
 *
 * UNITED STATES GOVERNMENT LEGENDS:  The portions of this Driver developed by 
 * 3Com, and the tc990image ("Software") is commercial in nature and developed 
 * solely at private expense.  The Software is delivered as 
 * "Commercial Computer Software" as defined in DFARS 252.227-7014 (June 1995) 
 * or as a commercial item as defined in FAR 2.101(a) and as such is provided 
 * with only such rights as are provided herein.  Technical data is provided 
 * with limited rights only as provided in DFAR 252.227-7015 (Nov. 1995) or 
 * FAR 52.227-14 (June 1987), whichever is applicable.
 *
 * TERM AND TERMINATION:  The licenses to the Software granted hereunder are 
 * perpetual unless terminated earlier as specified below or otherwise 
 * provided. You may terminate the licenses and this Agreement at any time by 
 * destroying the Software and Documentation together with all copies and 
 * merged portions in any form. The licenses and this Agreement will also 
 * terminate immediately if you fail to comply with any term or condition of 
 * this Agreement. Upon such termination you agree to destroy the Software and 
 * Documentation, together with all copies and merged portions in any form.
 *
 * LIMITED WARRANTIES AND LIMITATION OF LIABILITY: This driver and  tc990image 
 * are provided with no warranty, including warranty of fitness for any 
 * particular purpose.  3Com assumes no liability for any damages or injuries 
 * resulting from use of the driver or tc990image file.
 * GOVERNING LAW:   This License shall be governed by the laws of the State of 
 * California, U.S.A. excluding its conflicts of laws principles and excluding 
 * the United Nations Convention on Contracts for the 
 * International Sale of Goods.
 * SEVERABILITY:  In the event any provision of this License is found to be 
 * invalid, illegal or unenforceable, the validity, legality and enforceability
 * of any of the remaining provisions shall not in any way be affected or 
 * impaired and a valid, legal and enforceable provision of similar intent and 
 * economic impact shall be substituted therefor.
 * ENTIRE AGREEMENT:  This Agreement sets forth the entire understanding and 
 * agreement between you and 3Com and supersedes all prior agreements, whether 
 * written or oral, with respect to the Software and Documentation, and may be 
 * amended only in a writing signed by both parties.
 * Should you have any questions concerning this Agreement or if you desire to 
 * contact 3Com for any reason, please contact the 3Com subsidiary serving your
 * country, or write: 
 * 3Com Corporation, Customer Support Information, 
 * 5400 Bayfront Plaza, Santa Clara, CA  95052
 */

static const char *version = "3Com  3c990.c  v1.0.0b  10/2000  \n";

#if !defined(__OPTIMIZE__)  ||  !defined(__KERNEL__)
#warning  You must compile this file with the correct options!
#warning  See the last lines of the source file.
#error You must compile this driver with "-O".
#endif

#define UPDATE_INDEX( index, size, entries ) { index += size; if ( index == (entries * size ) )  index = 0; }

/*
 * Include files to support kernel versions 2.0.x and 2.2.x
 */
#include <linux/config.h>
#include <linux/version.h>
#ifdef MODULE

#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif
#include <linux/module.h>
#else
#define MOD_INC_USE_COUNT
#define MOD_DEC_USE_COUNT
#endif

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <linux/errno.h>
#include <linux/ioport.h>
#include <linux/slab.h>
#include <linux/interrupt.h>
#include <linux/pci.h>
#include <linux/init.h>
#include <asm/bitops.h>
#include <asm/io.h>

#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>
#include <linux/delay.h>    

#ifdef SMP
#include <linux/spinlock.h>
#endif

/*
 * B NIC Extension support
 */
//#define NICE_SUPPORT 1
#ifdef NICE_SUPPORT
#include "nicext.h"
#endif  /* NICE_SUPPORT */

#if (LINUX_VERSION_CODE >= 0x20100)
char kernel_version[] = UTS_RELEASE;
#else
#ifndef __alpha__
#define ioremap vremap
#define iounmap vfree
#endif
#endif

#define POLYNOMIAL                       0x04c11db7 
#define TC990_DOWNLOAD_BLOCK_SIZE        80

#if defined(MODULE) && LINUX_VERSION_CODE > 0x20115

#define MAX_UNITS 8

MODULE_AUTHOR("3Com <linux_support@3com.com>");
MODULE_DESCRIPTION("Driver for 3Com EtherLink 10/100 PCI NIC with 3XP Processor");

MODULE_PARM(min_pci_latency, "1-" __MODULE_STRING(8) "i");
MODULE_PARM(force, "1-" __MODULE_STRING(8) "i");

static u32 tc990_index = 0;
static u32 force[MAX_UNITS] = { 4, 4, 4, 4, 4, 4, 4, 4};
static u32 min_pci_latency[MAX_UNITS] = {64,64,64,64,64,64,64,64};
#endif

#if (LINUX_VERSION_CODE < 0x02030e)
#define net_device device
#endif

#if (LINUX_VERSION_CODE >= 0x02031b)
#define NEW_NETINIT
#endif

/*
 * Use the IO mode to access the registers
 */
#define USE_IO_MODE

#ifdef USE_IO_MODE
#define TC990_BASE_ADDRESS_REG    0
#define tc990_writel         outl
#define tc990_readl        inl
#else
#define TC990_BASE_ADDRESS_REG 1
#define tc990_writel        writel
#define tc990_readl        readl
#endif

#if (LINUX_VERSION_CODE < 0x020400)
#define DMA_ADDR u32

static inline void *pci_alloc_consistent(struct pci_dev *hwdev, size_t size,
                     DMA_ADDR *dma_handle)
{
    void *virt_ptr;

    virt_ptr = kmalloc(size, GFP_KERNEL);
    *dma_handle = virt_to_bus(virt_ptr);
    return virt_ptr;

}

#define pci_free_consistent(cookie, size, ptr, dma_ptr)    kfree(ptr)
#define pci_map_single(cookie, address, size, dir)    virt_to_bus(address)
#define pci_unmap_single(cookie, address, size, dir)
#else  // (LINUX_VERSION_CODE < 0x020400)

#define DMA_ADDR dma_addr_t
#endif // (LINUX_VERSION_CODE < 0x020400)

#if (LINUX_VERSION_CODE < 0x02032b)
/*
 * SoftNet
 */
#define dev_kfree_skb_irq(a)    dev_kfree_skb(a)
#define netif_wake_queue(dev)    clear_bit(0, &dev->tbusy)

static inline void netif_start_queue(struct net_device *dev)
{
    dev->tbusy = 0;
    dev->interrupt = 0;
    dev->start = 1;
}
static inline void netif_stop_queue(struct net_device *dev) 
{
    dev->tbusy = 1;
    dev->start = 0;
}

#define netif_queue_stopped(dev)    dev->tbusy
#define netif_running(dev)        dev->start
#else
#define NET_BH            0
#define tc990_mark_net_bh(foo)    {do{} while(0);}
#define tc990_if_down(dev)    {do{} while(0);}
#endif

#if LINUX_VERSION_CODE < 0x20123
#define test_and_set_bit(val, addr) set_bit(val, addr)
#endif
#if LINUX_VERSION_CODE <= 0x20139
#define    net_device_stats enet_statistics
#else
#define NETSTATS_VER2
#endif
#if LINUX_VERSION_CODE < 0x20155
#include <linux/bios32.h>
#define PCI_SUPPORT_VER1
#else
#define PCI_SUPPORT_VER2
#endif
#if LINUX_VERSION_CODE < 0x20159
#define dev_free_skb(skb) dev_kfree_skb(skb, FREE_WRITE);
#else
#define dev_free_skb(skb) dev_kfree_skb(skb);
#endif
#if ! defined(CAP_NET_ADMIN)
#define capable(CAP_XXX) (suser())
#endif

#define tc990_IMAGE_SIZE 0x10000
#include "3c990img.h"

#define tc990_VENDOR_ID 0x10B7

#define TX_ENTRIES        128 
#define RX_ENTRIES        128 
#define CMD_ENTRIES        64
#define RESPONSE_ENTRIES   64

#define NIC_STATUS_SUCCESS		0x1
#define NIC_STATUS_FAILURE		0x2

/*
 * 3cr990 commands
 */
#define tc990_CMD_TX_ENABLE             0x1
#define tc990_CMD_TX_DISABLE            0x2
#define tc990_CMD_RX_ENABLE             0x3
#define tc990_CMD_RX_DISABLE            0x4
#define tc990_CMD_RX_FILT_WRITE         0x5
#define tc990_CMD_READ_STATS            0x7
#define tc990_CMD_XCVR_SELECT           0x13
#define tc990_CMD_MAX_PKT_SIZE_WRITE    0x1A
#define tc990_CMD_PHYSICAL_MNGT_READ	0x1F
#define tc990_CMD_PHYSICAL_MNGT_WRITE	0x20
#define tc990_CMD_MCAST_HASH_MASK_WRITE 0x25
#define tc990_CMD_STATION_ADR_WRITE     0x26
#define tc990_CMD_STATION_ADR_READ      0x27

#define tc990_CMD_ERROR_TIMEOUT		    0xFF

/*
 * Descriptor type
 */
#define CMD_DSC_TYPE_CMD_FRAME       0x2
#define CMD_DSC_FRAME_VALID          ( 1 << 7 ) 
#define CMD_DSC_RESPONSE_NEEDED      ( 1 << 6 )
#define CMD_DSC_RESPONSE_NOT_NEEDED  0 

#define RESPONSE_DSC_ERROR_SET       ( 1 << 6 )
#define tc990_STATUS_CMD_FOUND         2

/*
 * 3cr990 registers
 */
#define tc990_SOFT_RESET_REG         0x0
#define tc990_INT_STATUS_REG         0x4
#define tc990_INT_ENABLE_REG         0x8
#define tc990_INT_MASK_REG           0xC
#define tc990_HostTo3XP_COMM_5_REG   0x1C
#define tc990_HostTo3XP_COMM_4_REG   0x20
#define tc990_HostTo3XP_COMM_3_REG   0x24
#define tc990_HostTo3XP_COMM_2_REG   0x28
#define tc990_HostTo3XP_COMM_1_REG   0x2C
#define tc990_HostTo3XP_COMM_0_REG   0x30
#define tc990_3XP2HOST_COMM_0_REG    0x40

#define tc990_MASK_ALL_INT        0xFFFFFFFF
#define tc990_UNMASK_ALL_INT      0x0
#define tc990_ENABLE_ALL_INT      0xFFFFFFEF
#define tc990_RESET_ALL           0x7F

/*
 * 3cr990 commands for handshake with firmware
 */
#define tc990_NULL_CMD                  0x00
#define tc990_BOOTCMD_REG_BOOT_RECORD   0x0FF
#define tc990_BOOTCMD_RUNTIME_IMAGE     0xFD
#define tc990_BOOTCMD_DOWNLOAD_COMPLETE 0xFB
#define tc990_BOOTCMD_SEGMENT_AVAILABLE 0xFC

/*
 * Waiting for boot signature
 */
#define tc990_WAITING_FOR_BOOT          0x7
#define tc990_WAITING_FOR_HOST_REQUEST  0xD
#define tc990_WAITING_FOR_SEGMENT       0x10

#define tc990_3XP_COMM_INT0  0x2

/*
 * Resource allocations
 */
#define TC990_SHARED_MEMORY_ALLOCATED           0x1
#define TC990_SLOT_MEMORY_ALLOCATED             0x2
#define TC990_BASE_ADDR_RESERVED                0x4
#define TC990_IO_MAPPED                         0x8
#define TC990_BASE_MEMORY_RESERVED              0x10
#define TC990_IRQ_RESERVED                      0x20
#define TC990_DEVICE_REGISTERED                 0x40
#define TC990_DEVICE_IN_ROOT_CHAIN              0x80
/*
 * Delay in microseconds to wait for the reset to be over
 */
#define tc990_WAIT_COUNTER        100000

#define ETHERNET_MAXIMUM_FRAME_SIZE 1514

#define tc990_RX_FILT_DIRECTED          0x1
#define tc990_RX_FILT_ALL_MULTICAST     0x2
#define tc990_RX_FILT_BROADCAST         0x4
#define tc990_RX_FILT_PROMISCUOUS       0x8
#define tc990_RX_FILT_HASH_MULTICAST    0x10

#define tc990_MULTICAST_BITS               0x3f

#ifdef DEBUG
extern u32 debug;
#endif

#define DEBUG_INITIALIZE	0x00000001
#define DEBUG_FUNCTION		0x00000002
#define DEBUG_IOCTL		0x00000004
#define DEBUG_CMD_RSP_RING	0x00000008
#define DEBUG_SEND		0x00000010
#define DEBUG_RECEIVE		0x00000020
#define DEBUG_INTERRUPT		0x00000040
#define DEBUG_ERROR		0x80000000

#ifdef DEBUG

#define DBGPRINT_INITIALIZE(A) if (debug & DEBUG_INITIALIZE) printk A
#define DBGPRINT_FUNCTION(A) if (debug & DEBUG_FUNCTION) printk A
#define DBGPRINT_SEND(A) if (debug & DEBUG_SEND) printk A
#define DBGPRINT_RECEIVE(A) if (debug & DEBUG_RECEIVE) printk A
#define DBGPRINT_INTERRUPT(A) if (debug & DEBUG_INTERRUPT) printk A
#define DBGPRINT_CMD_RSP_RING(A) if (debug & DEBUG_CMD_RSP_RING) printk A
#define DBGPRINT_IOCTL(A) if (debug & DEBUG_IOCTL) printk A
#define DBGPRINT_ERROR(A) printk A
#define DBGPRINT_INIT(A) printk A

#else

#define DBGPRINT_INITIALIZE(A)
#define DBGPRINT_FUNCTION(A)
#define DBGPRINT_SEND(A)
#define DBGPRINT_RECEIVE(A)
#define DBGPRINT_INTERRUPT(A)
#define DBGPRINT_CMD_RSP_RING(A)
#define DBGPRINT_QUERY(A)
#define DBGPRINT_SET(A)
#define DBGPRINT_IOCTL(A)
#define DBGPRINT_ERROR(A)
#define DBGPRINT_INIT(A)

#endif


/*
 * Slot information
 */
struct SLOT {

    u32 physical_addr_lo;
    u32 physical_addr_hi;
    u32 buffer_length;
    struct sk_buff* skb;
};


typedef struct _tc990_RING {

    void* RingBase;          /* ring address (sharedmemory) */
    u32   NoEntries;         /* number of entries in ring */
    u32   LastWriteUL;       /* send: last insert */

} tc990_RING;

typedef struct _TX_RING {
    void* RingBase;          /* ring address (sharedmemory) */
    u32   NoEntries;         /* number of entries in ring */
    u32   LastWriteUL;       /* send: last insert */
    u32   LastReadUL;        /* cleanup: last read - current read */
    u32   WriteRegister;     /* register used to single NIC */
    u32   PacketPendingNo;   /* Number of packets sent - not completed */
} TX_RING;

/*
 * This structure is updated by the driver and reaad by the f/w
 */
typedef struct _HOST_WRITE_INDEXES {
    volatile u32 regRxHiReadUL;
    volatile u32 regRxLoReadUL;
    volatile u32 regRxBuffWriteUL;
    volatile u32 regRespReadUL;
} HOST_WRITE_INDEXES;

/*
 * This structure is updated by the f/w and read by the driver
 */
typedef struct _HOST_READ_INDEXES {
    volatile u32 regTxLoReadUL;
    volatile u32 regTxHiReadUL;
    volatile u32 regRxLoWriteUL;
    volatile u32 regRxBuffReadUL;
    volatile u32 regCmdReadUL;
    volatile u32 regRespWriteUL;
    volatile u32 regRxHiWriteUL;
} HOST_READ_INDEXES;

/*
 * Main variable structure
 */
typedef struct _HOST_VAR_S {
    HOST_WRITE_INDEXES hvWriteS;
    HOST_READ_INDEXES hvReadS;
} HOST_VAR_S;

typedef struct _HOST_INIT_S {
    u32 hostVarsPS;     /* points to host variable data struct */
    u32 hostVarsHiUL;
    u32 hostTxLoStartUL;         /* Tx Lo priority ring */
    u32 hostTxLoStartHiUL;
    u32 hostTxLoSizeUL;
    u32 hostTxHiStartUL;         /* Tx Hi priority ring */
    u32 hostTxHiStartHiUL;
    u32 hostTxHiSizeUL;
    u32 hostRxLoStartUL;         /* Rx Lo priority ring */
    u32 hostRxLoStartHiUL;
    u32 hostRxLoSizeUL;
    u32 hostRxFreeStartUL;       /* Rx free buffer ring */
    u32 hostRxFreeStartHiUL;
    u32 hostRxFreeSizeUL;
    u32 hostCtrlStartUL;         /* Comand ring */
    u32 hostCtrlStartHiUL;
    u32 hostCtrlSizeUL;
    u32 hostRespStartUL;         /* Command Response ring */
    u32 hostRespStartHiUL;
    u32 hostRespSizeUL;
    u32 hostZeroWordUL;          /* (lo) physical address of zero word */
    u32 hostZeroWordHiUL;        /* (Hi) used for dma */
    u32 hostRxHiStartUL;         /* Rx Hi priority ring */
    u32 hostRxHiStartHiUL;
    u32 hostRxHiSizeUL;
} HOST_INIT_S;

/*
 * Receive descriptor. This structure is updated by the f/w
 */
typedef struct _RX_DSC {

    u32 Flags:8;
    u32 num_descriptor:8;
    u32 FrameLength:16;
    u32 virtual_addr_lo;
    u32 virtual_addr_hi;
    u32 RxStatus;
    u16 FilterResults;
    u16 res1;
    u32 res2;

} RX_DSC;

/*
 * Receive free descriptor. This structure is filled by the driver to
 * give a buffer to the firmware
 */

typedef struct _RX_FREE_DSC {

    u32 physical_addr_lo;
    u32 physical_addr_hi;
    u32 virtual_addr_lo;
    u32 virtual_addr_hi;

} RX_FREE_DSC;
/*
 * Usef for both commands and responses
 */

typedef struct {

    u32 Flags:8;
    u32 num_descriptor:8;
    u32 Command:16;
    u32 SequenceNo:16;
    u32 Parameter1:16;
    u32 Parameter2;
    u32 Parameter3;

} CMD_DSC;
/*
 * Stats response labels 
 */
typedef struct {
    u32 reserved;
    u32 reservedB;
    u32 pkt;
    u32 byte;
}rtnTx;

typedef struct {
    u32 reserved;
    u32 deferred;
    u32 lateCollision;
    u32 collisions;
}rtnTxMore;

typedef struct {
    u32 carrierLost;
    u32 multCollision;
    u32 reserved;
    u32 reservedB;
}rtnTxCrit;

typedef struct {
    u32 reserved;
    u32 txFiltered;
    u32 rxPkt;
    u32 rxByte;
}rtnTxRx;

typedef struct {
    u32 reserved;
    u32 fifoOverrun;
    u32 badSSD;
    u32 crcError;
}rtnRx;

typedef struct {
    u32 oversize;
    u32 reserved;
    u32 reservedB;
    u32 reservedC;
}rtnRxGeneral;

typedef struct {
    u32 filtered;
    u32 reserved;
    u32 reservedB;
    u32 reservedC;
}rtnRxEtc;

typedef union {
    CMD_DSC nrml;                 /* "normal" response values */
    rtnTx tx;
    rtnTxMore txMore;
    rtnTxCrit txCrit;
    rtnTxRx txRx;
    rtnRx rx;
    rtnRxGeneral rxGeneral;
    rtnRxEtc rxEtc;
} RESPONSE_DSC;

/* individual descriptors for each Fragment of a packet */

typedef struct _TX_DESC
{
    u8 fragFlagsUC;
    u8 fragReservedUC;
    u16 fragLenUW;

    u32 fragHostAddrLoUL;
    u32    fragHostAddrHiUL;
    u32 fragSpareUL;

} TX_DESC;

typedef struct
{
    u8 frFlagsUC;
    u8 frNumDescUC;            /* No of fragments in this packet */
    u16 frPktLen;

} TX_PKT_FLAGS;

typedef struct _TX_FRAME_DESC
{
      union
      {
            TX_PKT_FLAGS frFlagsS;
            u32 frFlagsUL;
      } frU;

    u32    pktPtrLoUL;            
    u32 pktPtrHiUL;
    u32 frProcFlagsUL;
} TX_FRAME_DESC;

/* bit defines for frFlagsUC */

#define FRAME_TYPE_FRAG_HDR 0x00    /* bottom 2 bits describe pkt */
#define FRAME_TYPE_PKT_HDR  0x01    /* bottom 2 bits describe pkt */

#define HOST_DESC_VALID     0x80    /* set when this pkt contains data */

typedef struct _tc990_FILE_HEADER {
    u8 tagID[8];
    u32 Version;
    u32 NumSections;
    u32 ExecuteAddress;
} tc990_FILE_HEADER; 

typedef struct _tc990_FILE_SECTION {

    u32 num_bytes;
    u16 checksum;
    u16 reserved;
    u32 start_address;

} tc990_FILE_SECTION;

/**********************************************************************/
// Device IDs for top 16bits of PCI_CFG_ID reg 
/**********************************************************************/
#define DEV_ID_CLIENT_NO_DES   0x9900
#define DEV_ID_CLIENT_DES      0x9902
#define DEV_ID_CLIENT_3DES     0x9903

#define DEV_ID_SERVER_NO_DES   0x990A
#define DEV_ID_SERVER_DES      0x9908
#define DEV_ID_SERVER_3DES     0x9909

#define DEV_ID_T2              0x9904
#define DEV_ID_TYPHOON         0x9905
                                     
/**********************************************************************/
// Subsystem ID's 
/**********************************************************************/
#define SUBSYSTEM_ID_CLIENT             0x1000
#define SUBSYSTEM_ID_SERVER             0x2000
#define SUBSYSTEM_ID_MASK               0xf000

#define SUBSYSTEM_COPPER_PHY            0x0000
#define SUBSYSTEM_FIBER_PHY             0x0100
#define SUBSYSTEM_FIBER_MASK            0x0100

#define SUBSYSTEM_STANDARD_FIRMWARE     0x0000
#define SUBSYSTEM_SECURED_FIRMWARE      0x0010
#define SUBSYSTEM_FIRMWARE_MASK         0x0010

#define SUBSYSTEM_VARIABLE_DES_MONGOOSE 0x0000
#define SUBSYSTEM_SINGLE_DES_SIDEWINDER 0x0001
#define SUBSYSTEM_TRIPLE_DES_SIDEWINDER 0x0002
#define SUBSYSTEM_SUBSYSTEM_DES_MASK    0x0003

struct pci_id_info {
    const char *name;
    u16 device_id;
    u16 subsystem_id;
};

static struct pci_id_info pci_tbl[] = {
    //Cards using old Device ID and sub-system ID dydtem
	{"3Com 10/100 PCI NIC w/3XP (3C990-TX)",           0x9900,0x9900},
    {"3Com 10/100 PCI NIC w/3XP (3CR990-TX-95)",       0x9902,0x9902},
    {"3Com 10/100 PCI NIC w/3XP (3CR990-TX-97)",       0x9903,0x9903},
    {"3Com 10/100 PCI Server NIC w/3XP (3C990SVR)",    0x990A,0x990A},
    {"3Com 10/100 PCI Server NIC w/3XP (3CR990SVR95)", 0x9908,0x9908},
    {"3Com 10/100 PCI Server NIC w/3XP (3CR990SVR97)", 0x9909,0x9909},

    //Cards using the new system
    {"3Com 10/100 PCI NIC w/3XP (3C990B-TX-M)",		       0x9904,0x1000},
    {"3Com 10/100 PCI Server NIC w/3XP (3C990BSVR)",       0x9904,0x2000},
    {"3Com 100 PCI Fiber NIC w/3XP (3CR990-FX-95)",        0x9905,0x1101},
    {"3Com 100 PCI Fiber NIC w/3XP (3CR990-FX-97)",        0x9905,0x1102},
    {"3Com 100 PCI Fiber Server NIC w/3XP (3CR990-FX-95)", 0x9905,0x2101},
    {"3Com 100 PCI Fiber Server NIC w/3XP (3CR990-FX-97)", 0x9905,0x2102},
    {0,0,0},                  /* 0 terminated list. */
};

struct tc990_shared {

    HOST_INIT_S init; 
    u32 slack;
    HOST_VAR_S var; 
    RX_FREE_DSC rxFree[RX_ENTRIES];
    RX_DSC rxLo[RX_ENTRIES], rxHi[RX_ENTRIES]; 
    CMD_DSC cmd[CMD_ENTRIES];
    RESPONSE_DSC rsp[RESPONSE_ENTRIES];

};

struct tc990_ring_info {

    tc990_RING RxHiRing, RxLoRing, RxBuffRing;  /* receive information */
    TX_RING TxHiRing, TxLoRing;             /* send information */
    tc990_RING CmdRing, RspRing;            /* command ring information */

};

struct tc990_private { 

    struct tc990_shared *shared;   
    u32 shared_size;
    DMA_ADDR shared_dma_addr;
 
    struct tc990_ring_info ring_info;     

    u8 *DownloadPageVirtual;
    DMA_ADDR DownloadPagePhysical;
    u32 ringPhysical;
    u32 BufferPhysical;
    struct SLOT* slot;
    u32 slot_memory_size;
    DMA_ADDR slot_physical;

    struct net_device *next_module;    /* Link for devices of this type. */
    const char *product_name;
    struct net_device_stats stats;
    u16 chip_id;
    unsigned char pci_bus, pci_devfn;
    u8 tx_full;        
    spinlock_t lock; 
#if LINUX_VERSION_CODE >= 0x20343
    wait_queue_head_t wait_q;
#else
    struct wait_queue *wait_q;   
#endif
    struct pci_dev *pdev;
    u32 index;
    u32 resources_reserved;
    u8 StationAddress[6];
    unsigned long waiting_for_response;
#ifdef NICE_SUPPORT
	void (*nice_rx)( struct sk_buff*, void* );
	void* nice_ctx;
#endif  /* NICE_SUPPORT */


};

/*----------------------------------------------------------------------
 * rebuild_address --
 *  This Inline Function is called to reconstruct an addess from 2 ints.
 *	Note: this only returns the LowBits in a 32 Bit System
 *
 * Arguments:
 *  HighBits     - The Upper portion of the Address

 *  LowBits      - The Lower portion of the Address
 *
 * Results:
 *  Pointer recontructed from the 2 intsnone
 *----------------------------------------------------------------------
 */
inline void* rebuild_address(unsigned long HighBits, unsigned long LowBits)    
{
#if (BITS_PER_LONG == 64)
	return (void*)((HighBits << 32) | (LowBits));
#else
	return (void*)LowBits;
#endif
}
/*----------------------------------------------------------------------
 * upper_bits --
 *  This Inline Function is called to return the upper 32 bits of an addess
 *	Note: this returns 0 on a 32 Bit System
 *
 * Arguments:
 *  address     - The address to get the upper bits from
 *
 * Results:
 *  32 bit number containing the upper bits of the address
 *----------------------------------------------------------------------
 */
inline u32 upper_bits(void* address)    
{
#if (BITS_PER_LONG == 64)
	return (((unsigned long)address) >> 32);
#else
	return 0;
#endif
}

static void update_tx_ring_index (u32 *writePUW);
static struct net_device *tc990probe( struct net_device *dev, long ioaddr, u16 irq, u16 chp_idx, struct pci_dev* pdev);

static int  tc990_open(struct net_device *dev);                    
static void tc990_interrupt(int irq, void *dev_inst, struct pt_regs *regs);  
static int  tc990_close(struct net_device *dev);                   
static int  tc990_start_tx(struct sk_buff *skb, struct net_device *dev);  
static struct net_device_stats *tc990_get_stats(struct net_device *dev);  

static void stats_handler(struct tc990_private *np);
static void release_buf_to_firmware( struct net_device *dev, u32 slot_index );
static void process_receive(struct net_device *dev, tc990_RING *pRxRing,
volatile u32 *regRxReadUL, volatile u32 *regRxWriteUL);

static u16 calculate_buffer_checksum(u16 *Buffer,  u32 Count);
static u16 download_boot_record(struct net_device *dev, u16 process);
static u16 download_runtime_image(struct net_device *dev);
static u16 init_ring_zone( struct net_device *dev ) ;
static u16 issue_command( struct net_device *dev, u16 command, u16 parameter1, 
u32 parameter2, u32 parameter3, u16 *return1, u32 *return2, u32 *return3, 
u8 wait_for_response);

static u8 process_response( struct tc990_private *np, u32 response_read_index, u16 command );

static void free_resources(struct net_device *dev);

static void process_send_complete( struct net_device *dev, TX_RING *pTxRing,
    volatile u32 *xmitReadIndex)    ;
static void tc990_set_rx_mode(struct net_device *dev);


int tc990_init_module(void);
void tc990_cleanup_module(void);

int SetMacAddress(struct net_device *Device, void *SockPtr);
u32 WriteMacAddressToNic(struct net_device *dev);
u32 ReadPermanentAddress(struct net_device *dev);
int IoctlHandler( struct net_device *dev, struct ifreq *ifr, int command );
