/*
 *  linux/fs/supermount/revalidate.c
 *
 *  Original version:
 *      Copyright (C) 1995
 *      Stephen Tweedie (sct@dcs.ed.ac.uk)
 *
 *  Rewriten for kernel 2.2 & 2.4. (C) 1999, 2000 Alexis Mikhailov
 *                                    (alexis@abc.cap.ru)
 *
 */

#include <linux/errno.h>
#include <linux/fs.h>

#include "supermount_i.h"

static int
supermount_dentry_revalidate(struct dentry *dentry, int flags)
{
	struct dentry *subd;
	int rc = 1;

	if (subfs_go_online(dentry->d_sb))
		goto bad_dentry;
	spin_lock(&dcache_lock);
	if (dentry->d_inode && is_inode_obsolete(dentry->d_inode)) {
		spin_unlock(&dcache_lock);
		goto bad_dentry;
	}
	spin_unlock(&dcache_lock);
	subd = get_subfs_dentry(dentry);
	if (IS_ERR(subd))
		goto bad_dentry;

	if (subd->d_op && subd->d_op->d_revalidate)
		rc = subd->d_op->d_revalidate(subd, flags);

	dput(subd);
out:
	subfs_go_inactive(dentry->d_sb);
	return rc;
bad_dentry:
	rc = 0;
	goto out;
}

struct dentry_operations supermount_dir_dops = {
	.d_revalidate	= supermount_dentry_revalidate,
};

struct dentry_operations supermount_file_dops = {
	.d_revalidate	= supermount_dentry_revalidate,
};
