#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/file.h>
#include <linux/ipc.h>
#include <linux/grsecurity.h>

#include <asm/uaccess.h>

/*
   initialize variables  These initial states are used in two situations:

   1) certain features of grsecurity are disabled (if they are enabled, 
      then they are handled below)
   2) the sysctl feature is enabled, in which case everything is disabled
      at bootup and users must echo values to /proc to set the features
*/

int grsec_enable_link;
int grsec_enable_dmesg;
int grsec_enable_fifo;
int grsec_enable_execve;
int grsec_enable_execlog;
int grsec_enable_signal;
int grsec_enable_forkfail;
int grsec_enable_time;
int grsec_enable_group;
int grsec_audit_gid;
int grsec_enable_chdir;
int grsec_enable_audit_ipc;
int grsec_enable_mount;
int grsec_enable_chroot_findtask;
int grsec_enable_chroot_mount;
int grsec_enable_chroot_shmat;
int grsec_enable_chroot_fchdir;
int grsec_enable_chroot_double;
int grsec_enable_chroot_pivot;
int grsec_enable_chroot_chdir;
int grsec_enable_chroot_chmod;
int grsec_enable_chroot_mknod;
int grsec_enable_chroot_nice;
int grsec_enable_chroot_execlog;
int grsec_enable_chroot_caps;
int grsec_enable_tpe;
int grsec_tpe_gid;
int grsec_enable_tpe_all;
int grsec_enable_randpid;
int grsec_enable_randid;
int grsec_enable_randsrc;
int grsec_enable_randrpc;
int grsec_enable_randping;
int grsec_enable_socket_all;
int grsec_socket_all_gid;
int grsec_enable_socket_client;
int grsec_socket_client_gid;
int grsec_enable_socket_server;
int grsec_socket_server_gid;
int grsec_lock;

/* 
   handle the variables if parts of grsecurity are configured without sysctl 
   we do this as a function since it runs on bootup
*/

void grsecurity_init(void)
{
grsec_lock = 1;
#ifdef CONFIG_GRKERNSEC_AUDIT_GROUP
grsec_enable_group = 1;
grsec_audit_gid = CONFIG_GRKERNSEC_AUDIT_GID;
#endif
#ifdef CONFIG_GRKERNSEC_AUDIT_CHDIR
grsec_enable_chdir = 1;
#endif
#ifdef CONFIG_GRKERNSEC_AUDIT_IPC
grsec_enable_audit_ipc = 1;
#endif
#ifdef CONFIG_GRKERNSEC_AUDIT_MOUNT
grsec_enable_mount = 1;
#endif
#ifdef CONFIG_GRKERNSEC_LINK
grsec_enable_link = 1;
#endif
#ifdef CONFIG_GRKERNSEC_DMESG
grsec_enable_dmesg = 1;
#endif
#ifdef CONFIG_GRKERNSEC_FIFO
grsec_enable_fifo = 1;
#endif
#ifdef CONFIG_GRKERNSEC_EXECVE
grsec_enable_execve = 1;
#endif
#ifdef CONFIG_GRKERNSEC_EXECLOG
grsec_enable_execlog = 1;
#endif
#ifdef CONFIG_GRKERNSEC_SIGNAL
grsec_enable_signal = 1;
#endif
#ifdef CONFIG_GRKERNSEC_FORKFAIL
grsec_enable_forkfail = 1;
#endif
#ifdef CONFIG_GRKERNSEC_TIME
grsec_enable_time = 1;
#endif
#ifdef CONFIG_GRKERNSEC_CHROOT_FINDTASK
grsec_enable_chroot_findtask = 1;
#endif
#ifdef CONFIG_GRKERNSEC_CHROOT_MOUNT
grsec_enable_chroot_mount = 1;
#endif
#ifdef CONFIG_GRKERNSEC_CHROOT_FCHDIR
grsec_enable_chroot_fchdir = 1;
#endif
#ifdef CONFIG_GRKERNSEC_CHROOT_SHMAT
grsec_enable_chroot_shmat = 1;
#endif
#ifdef CONFIG_GRKERNSEC_CHROOT_DOUBLE
grsec_enable_chroot_double = 1;
#endif
#ifdef CONFIG_GRKERNSEC_CHROOT_PIVOT
grsec_enable_chroot_pivot = 1;
#endif
#ifdef CONFIG_GRKERNSEC_CHROOT_CHDIR
grsec_enable_chroot_chdir = 1;
#endif
#ifdef CONFIG_GRKERNSEC_CHROOT_CHMOD
grsec_enable_chroot_chmod = 1;
#endif
#ifdef CONFIG_GRKERNSEC_CHROOT_MKNOD
grsec_enable_chroot_mknod = 1;
#endif
#ifdef CONFIG_GRKERNSEC_CHROOT_NICE
grsec_enable_chroot_nice = 1;
#endif
#ifdef CONFIG_GRKERNSEC_CHROOT_EXECLOG
grsec_enable_chroot_execlog = 1;
#endif
#ifdef CONFIG_GRKERNSEC_CHROOT_CAPS
grsec_enable_chroot_caps = 1;
#endif
#ifdef CONFIG_GRKERNSEC_TPE
grsec_enable_tpe = 1;
grsec_tpe_gid= CONFIG_GRKERNSEC_TPE_GID;
#ifdef CONFIG_GRKERNSEC_TPE_ALL
grsec_enable_tpe_all = 1;
#endif
#endif
#ifdef CONFIG_GRKERNSEC_RANDPID
grsec_enable_randpid = 1;
#endif
#ifdef CONFIG_GRKERNSEC_RANDID
grsec_enable_randid = 1;
#endif
#ifdef CONFIG_GRKERNSEC_RANDSRC
grsec_enable_randsrc = 1;
#endif
#ifdef CONFIG_GRKERNSEC_RANDRPC
grsec_enable_randrpc = 1;
#endif
#ifdef CONFIG_GRKERNSEC_RANDPING
grsec_enable_randping = 1;
#endif
#ifdef CONFIG_GRKERNSEC_SOCKET_ALL
grsec_enable_socket_all = 1;
grsec_socket_all_gid= CONFIG_GRKERNSEC_SOCKET_ALL_GID;
#endif
#ifdef CONFIG_GRKERNSEC_SOCKET_CLIENT
grsec_enable_socket_client = 1;
grsec_socket_client_gid= CONFIG_GRKERNSEC_SOCKET_CLIENT_GID;
#endif
#ifdef CONFIG_GRKERNSEC_SOCKET_SERVER
grsec_enable_socket_server = 1;
grsec_socket_server_gid= CONFIG_GRKERNSEC_SOCKET_SERVER_GID;
#endif

return;
}

void gr_handle_exec_args(struct linux_binprm * bprm, char ** argv)
{
	char grarg[64] = {0};
	__u8 execlen = 0;
	unsigned int i;

	if (unlikely(!argv)) goto log;

	for(i = 0; i < bprm->argc && execlen < 62; i++) {
		char *p;
		__u8 len;

		if (get_user(p, argv + i))
			goto log;
		if (!p)
			goto log;
		len = strnlen_user(p, 62 - execlen);
		if (len > 62 - execlen)
			len = 62 - execlen;
		else if (len > 0)
			len--;
		if (copy_from_user(grarg + execlen, p, len))
			goto log;
		execlen += len;
		*(grarg + execlen) = ' ';
		*(grarg + execlen + 1) = '\0';
		execlen++;
	}

log:
	security_audit(GR_EXEC_AUDIT_MSG, 
			kdevname(bprm->file->f_dentry->d_inode->i_dev),
			bprm->file->f_dentry->d_inode->i_ino, grarg, 
			DEFAULTSECARGS);
	return;
}

int gr_tpe_allow(struct file * file)
{
	struct inode * inode = file->f_dentry->d_parent->d_inode;

	if(current->uid && grsec_enable_tpe && in_group_p(grsec_tpe_gid) && 
	   (inode->i_uid || (!inode->i_uid && ((inode->i_mode & S_IWGRP) ||
	   (inode->i_mode & S_IWOTH)))))
		return 0;

	return 1;
}

int gr_tpe_all_allow(struct file * file)
{
	struct inode * inode = file->f_dentry->d_parent->d_inode;

	if(current->uid && grsec_enable_tpe && grsec_enable_tpe_all && 
	   ((inode->i_uid && (inode->i_uid != current->uid)) ||
	   (inode->i_mode & S_IWGRP) || (inode->i_mode & S_IWOTH)))
		return 0;

	return 1;
}

int gr_pid_is_chrooted(struct task_struct *p)
{
	if(!grsec_enable_chroot_findtask) return 0;

	if(p && p->fs && p->fs->root && p->fs->root->d_inode &&
	   child_reaper && child_reaper->fs && child_reaper->fs->root &&
	   child_reaper->fs->root->d_inode && current && current->fs &&
	   current->fs->root && current->fs->root->d_inode) {
		if(proc_is_chrooted(current) && !have_same_root(current, p))
				return 1;
	}

	return 0;
}

int gr_chroot_fchdir(struct dentry *u_dentry, struct vfsmount *u_mnt)
{
	if (!grsec_enable_chroot_fchdir) return 1;

	if (!proc_is_chrooted(current))
		return 1;
	else {
		struct dentry *dentry = u_dentry;
		struct vfsmount *mnt = u_mnt;
		struct dentry *realroot;
		struct vfsmount *realrootmnt;
		struct dentry *currentroot;
		struct vfsmount *currentmnt;

		read_lock(&child_reaper->fs->lock);
		realrootmnt = mntget(child_reaper->fs->rootmnt);
		realroot = dget(child_reaper->fs->root);
		read_unlock(&child_reaper->fs->lock);

		read_lock(&current->fs->lock);
		currentmnt = mntget(current->fs->rootmnt);
		currentroot = dget(current->fs->root);
		read_unlock(&current->fs->lock);

		spin_lock(&dcache_lock);
		for(;;) {
			if (unlikely((dentry == realroot && mnt == realrootmnt) ||
				     (dentry == currentroot && mnt == currentmnt)))
				break;
			if (unlikely(dentry == mnt->mnt_root || IS_ROOT(dentry))) {
				if (mnt->mnt_parent == mnt)
					break;
				dentry = mnt->mnt_mountpoint;
				mnt = mnt->mnt_parent;
				continue;
			}
			dentry = dentry->d_parent;
		}
		spin_unlock(&dcache_lock);

		dput(currentroot);
		mntput(currentmnt);

		if (dentry == realroot && mnt == realrootmnt) {
		/* ok, they're definitely trying to fchdir outside of the
		   chroot. */
			dput(realroot);
			mntput(realrootmnt);
			return 0;
		} else {
			dput(realroot);
			mntput(realrootmnt);
			return 1;
		}
	}
}				

struct shmid_kernel
{
	struct kern_ipc_perm    shm_perm;
	struct file *           shm_file;
	int                     id;
	unsigned long           shm_nattch;
	unsigned long           shm_segsz;
	time_t                  shm_atim;
	time_t                  shm_dtim;
	time_t                  shm_ctim;
	pid_t                   shm_cprid;
	pid_t                   shm_lprid;

        /* added by grsecurity */
	time_t                  shm_createtime;
	pid_t                   shm_lapid;
};


int gr_chroot_shmat(const struct shmid_kernel *shp)
{
	struct task_struct *p, **htable;

	if (unlikely(!grsec_enable_chroot_shmat)) return 1;

	if (likely(!proc_is_chrooted(current)))
		return 1;

	read_lock(&tasklist_lock);

	htable = &pidhash[pid_hashfn(shp->shm_cprid)];

        for(p = *htable; p && p->pid != shp->shm_cprid; p = p->pidhash_next)
		;

	if (unlikely(p && !have_same_root(current, p) &&
		     (p->start_time < shp->shm_createtime))) {
		read_unlock(&tasklist_lock);
		return 0;
	}
	
	if (unlikely(!p)) {
		htable = &pidhash[pid_hashfn(shp->shm_lapid)];
		for(p = *htable; p && p->pid != shp->shm_lapid; p = p->pidhash_next)
			;

		if (unlikely(p && !have_same_root(current, p))) {
			read_unlock(&tasklist_lock);
			return 0;
		}
	}

	read_unlock(&tasklist_lock);

        return 1;
}

