#ifndef _LINUX_SWSUSP_H
#define _LINUX_SWSUSP_H

#if defined(SUSPEND_C) || defined(ACPI_C)
#include <asm/suspend.h>
#endif
#include <linux/swap.h>
#include <linux/notifier.h>
#include <linux/config.h>

extern unsigned char software_suspend_enabled;

#define NORESUME	 1
#define RESUME_SPECIFIED 2

#ifdef CONFIG_SOFTWARE_SUSPEND
extern int swsusp_state[5];

/* page backup entry */
typedef struct pbe {
	unsigned long address1;		/* original address of pageset1 entry */
	unsigned long address2;		/* original address of pageset2 entry */
	unsigned long address3;		/* copy back address */
	swp_entry_t swap_address1;
	swp_entry_t swap_address2;
	swp_entry_t dummy[3];		/* we need at least one sizeof(long) scratch space at 
					 * end of page (see link, diskpage) and
					 * sizeof(struct pbe) must be multiple of 4*sizeof(long)
					 */
} suspend_pagedir_t;

#define SWAP_FILENAME_MAXLENGTH	32

struct suspend_header {
	__u32 version_code;
	unsigned long num_physpages;
	char machine[65];
	char version[65];
	int num_cpus;
	int page_size;
	suspend_pagedir_t *suspend_pagedir;
	unsigned int pageset1_size;
	unsigned int pageset2_size;
	unsigned int pagedir_size;
	int param0;
	int param1;
	int param2;
	int param3;
	int param4;
	struct swap_location {
		char filename[SWAP_FILENAME_MAXLENGTH];
	} swap_location[MAX_SWAPFILES];
};

#define SUSPEND_PD_PAGES(x)     (((x)*sizeof(struct pbe))/PAGE_SIZE+1)
#define PAGEDIR_CAPACITY(x)     (((x)*PAGE_SIZE/sizeof(struct pbe)))
   
extern struct tq_struct suspend_tq;

/* mm/vmscan.c */
extern int shrink_mem(void);

/* kernel/suspend.c */
extern void software_suspend_pending(void);
extern void software_resume(void);

extern int register_suspend_notifier(struct notifier_block *);
extern int unregister_suspend_notifier(struct notifier_block *);
extern void refrigerator(unsigned long);
extern struct page * last_suspend_cache_page;

#ifdef CONFIG_MTRR
#define SOFTWARE_SUSPEND_MTRR 1
#else
#undef SOFTWARE_SUSPEND_MTRR
#endif

#else
#define software_suspend_pending()	do { } while(0)
#define software_resume()		do { } while(0)
#define register_suspend_notifier(a)	do { } while(0)
#define unregister_suspend_notifier(a)	do { } while(0)
#define refrigerator(a)			do { BUG(); } while(0)
#endif

#endif /* _LINUX_SWSUSP_H */
